from django.contrib.auth.backends import ModelBackend
from freeipa_auth.freeipa_utils import FreeIpaSession
from django.contrib.auth.models import Permission, Group
from django.contrib.auth import get_user_model
import requests
import logging

logger = logging.getLogger(__name__)

User = get_user_model()


class FreeIpaRpcAuthBackend(ModelBackend):
    """
    Free IPA RPC Authentication backend.
    Basic JSON RPC session authentication on free ipa servers
    with django user group sync.
    """

    def __init__(self):
        self.settings = FreeIpaAuthSettings()

    def authenticate(self, *args, **kwargs):
        """
        Overriden method of ModelBackend.
        Authenticate on freeipa server and sync django user groups.
        :param username:
        :param password:
        :param tries: Number of tries for authentication (internal use only)
        :return:
        """

        if self.settings.BACKEND_ENABLED:

            # get arguments from Django auth __init__
            username = kwargs.get('username', None)
            password = kwargs.get('password', None)
            tries = kwargs.get('tries', 1)

            # Grab freeipa server from settings
            server = self.settings.SERVER

            # if second try, allow for a single failover server only
            if tries == 2:
                server = self.settings.FAILOVER_SERVER

            # Specify ssl public cert for a mutual SSL handshake
            ssl_verify = self.settings.SSL_VERIFY

            # Setup FreeIPA user session
            user_session = FreeIpaSession(server, ssl_verify=ssl_verify)

            message = "Attempting to authenticate user on server: {server}"
            logger.info(message.format(server=server))

            try:
                # Authenticate and get response via RPC protocol
                response = user_session.authenticate(username, password)

                # Check response status code
                logged_in = response.status_code == 200

                # If credentials were valid then sync and return the user
                # Django will handle user sessions from here
                if logged_in:
                    if not self.is_authorized(user_session):
                        return
                    return self.update_user(user_session)

            except requests.ConnectionError:
                # If there was a connection error, we can try the
                # failover server once and return user
                logger.critical("Main FreeIPA server connection error")
                if tries == 1 and self.settings.FAILOVER_SERVER:
                    return self.authenticate(username, password, tries=tries+1)
                else:
                    raise

    def is_authorized(self, user_session):
        """
        If the AUTHORIZE_ALL_USERS setting is set to False
        then we check to see if freeipa user has groups0
        that are user flags to authorize basic login.
        :param user_session:
        :return:
        """
        if not self.settings.AUTHORIZE_ALL_USERS:
            # If not authorize all then we check to see if
            # user has any group flags to authorize
            groups = self.get_django_user_groups(user_session.groups)
            for group in groups:
                for flags in self.settings.USER_FLAGS_BY_GROUP.values():
                    if group in flags:
                        return True
            return False
        return True

    def update_user(self, user_session):
        """
        Sync freeipa user to django with user freeipa user groups groups
        :param user_session: freeipa_user_session obj
        :return:
        """

        user, created = User.objects.get_or_create(username=user_session.user)

        # Set random (secret) pass for freeipa user.
        # This user does not need to, and cannot, login
        # via classic django auth.
        user.set_password(User.objects.make_random_password(length=100))

        if not created and not self.settings.ALWAYS_UPDATE_USER:
            return user

        # Update user attrs
        self.update_user_attrs(user, user_session.user_data)

        # Sync freeipa user groups with current user
        django_user_groups = self.get_django_user_groups(user_session.groups)
        django_user_perms = self.get_django_user_perms(user_session.groups)
        self.update_user_groups(user, django_user_groups, django_user_perms)

        user.save()
        return user

    def update_user_attrs(self, user, user_session_data):

        for attr, key in self.settings.USER_ATTRS_MAP.items():
            attr_value = user_session_data[key]
            setattr(user, attr, attr_value.pop() if isinstance(attr_value, list) else attr_value)  # noqa: E501

    def update_user_groups(self, user, group_names, perm_codenames):
        """
        Add user to django groups and give user permissions
        based on freeipa permissions, also remove user from
        any groups that do not exist.
        :param user: Django user obj
        :param group_names: List of user group names
        :param perm_codenames: List of user permission codenames
        :return:
        """

        # Set flags on user attrs
        for group, flags in self.settings.USER_FLAGS_BY_GROUP.items():
            setattr(user, group, any(flag in group_names for flag in flags))

        # Update user groups
        if self.settings.UPDATE_USER_GROUPS:
            user.groups.add(*Group.objects.filter(name__in=group_names))
            # Remove now invalid user groups if any
            user.groups.remove(*user.groups.exclude(name__in=group_names))

        # Update user permissions
        if self.settings.UPDATE_USER_PERMISSIONS_BY_GROUP:
            # Allow adding one off permissions to select users.
            perms = Permission.objects.filter(codename__in=perm_codenames)
            user.user_permissions.add(*perms)
            # Remove now invalid user permissions if any
            old_perms = user.user_permissions.exclude(codename__in=perm_codenames)  # noqa: E501
            user.user_permissions.remove(*old_perms)

    def get_django_user_groups(self, groups):
        """
        Parses and returns django specific user groups from freeipa user groups
        :return:
        """
        freeipa_user_groups_prefix = self.settings.REQUIRE_GROUP_PREFIX
        if freeipa_user_groups_prefix:
            return [group.split(freeipa_user_groups_prefix).pop()
                    for group in groups if freeipa_user_groups_prefix in group]
        return groups

    def get_django_user_perms(self, groups):
        """
        Parses and returns django specific user permissions from
        freeipa user groups.
        :return:
        """
        user_permissions_prefix = self.settings.REQUIRE_PERMISSION_PREFIX
        if user_permissions_prefix:
            return [group.split(user_permissions_prefix).pop()
                    for group in groups if user_permissions_prefix in group]
        return groups


class FreeIpaAuthSettings(object):

    defaults = {
        'BACKEND_ENABLED': True,
        'SERVER': None,
        'FAILOVER_SERVER': None,
        'SSL_VERIFY': True,
        'UPDATE_USER_GROUPS': False,
        'UPDATE_USER_PERMISSIONS_BY_GROUP': False,
        'USER_FLAGS_BY_GROUP': {},
        'USER_ATTRS_MAP': {},
        'REQUIRE_GROUP_PREFIX': None,
        'REQUIRE_PERMISSION_PREFIX': None,
        'ALWAYS_UPDATE_USER': True,
        'AUTHORIZE_ALL_USERS': False,
    }

    def __init__(self, prefix='FREEIPA_AUTH_'):
        """
        Load FreeIPA Auth settings and set defaults
        if they do not exists
        """
        from django.conf import settings

        for name, default in self.defaults.items():
            value = getattr(settings, prefix + name, default)
            setattr(self, name, value)



import requests
import logging
import json


logger = logging.getLogger(__name__)


class FreeIpaSession(object):

    """FreeIPA session constructor for RPC authentication"""

    # Base login POST headers
    login_headers = {'referer': '',
                     'Content-Type': 'application/x-www-form-urlencoded',
                     'Accept': 'text/plain'}

    # Bse session POST headers
    session_headers = {'referer': '',
                       'Content-Type': 'application/json',
                       'Accept': 'application/json'}

    # Base user specific POST data
    user_post_data = {'item': [], 'method': 'user_show',
                      'params': {'all': True, 'raw': False}}

    # Base session POST data
    session_post_data = {'id': 0, 'method': '', 'params': []}

    def __init__(self, host_server, ssl_verify=False):

        self.host_server = host_server
        self.ssl_verify = ssl_verify
        self.user = None
        self.user_is_authenticated = False
        self.user_data = None
        self.session = requests.Session()

    def authenticate(self, user, password):
        """
        Authenticates user on freeipa backend and returns the session response
        :param user: string
        :param password: string
        :return: session response
        """
        url_template = 'https://{host_server}/ipa/session/login_password'
        ipa_login_url = url_template.format(host_server=self.host_server)

        # Add the referer header
        self.login_headers['referer'] = ipa_login_url

        # Set POST data
        login_data = {'user': user, 'password': password}

        logger.debug("User is attempting to authenticate via FreeIPA...")

        response = self.session.post(ipa_login_url,
                                     headers=self.login_headers,
                                     data=login_data,
                                     verify=self.ssl_verify)

        self.user = user
        # If user is authenticated, get user_data from the freeipa server
        if response.status_code == 200:
            logger.info("User successfully authenticated via FreeIPA")
            self.user_is_authenticated = True
            self.user_data = self._get_user_data()
        else:
            logger.info("User failed to authenticate via FreeIPA")

        return response

    def make_session_request(self, post_data):
        """
        Base POST request once user is authenticated
        and a session is established
        :param post_data: post_data to update base session post data
        :return:
        """

        url_template = 'https://{host_server}/ipa/session/json'
        ipa_session_url = url_template.format(host_server=self.host_server)

        # Set the referer header
        self.session_headers['referer'] = ipa_session_url

        # Update session POST data with user specific data
        self.session_post_data.update({'method': post_data['method'],
                                       'params': [
                                           post_data['item'],
                                           post_data['params']
                                       ]})

        debug_message = 'Making {method} request to {url}'
        logger.debug(debug_message.format(method=post_data['method'],
                                          url=ipa_session_url))

        request = self.session.post(
            ipa_session_url, headers=self.session_headers,
            data=json.dumps(self.session_post_data),
            verify=self.ssl_verify
        )

        results = request.json()
        return results

    def _get_user_data(self):
        """
        Internal method to grab user data on freeipa server upon authentication
        :return:
        """

        self.user_post_data['item'] = [self.user]

        if self.user_is_authenticated:
            response = self.make_session_request(self.user_post_data)
            return response['result']['result']

    @property
    def groups(self):
        """
        Returns all groups of which currently authenticated user is a member
        :return: List of groups
        """
        return self.user_data['memberof_group'] if self.user_data else []



import pytest
import getpass
from django.contrib.auth import get_user_model
from collections import namedtuple
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import Group
from django.conf import settings as django_settings


@pytest.fixture
def test_user(request, db):
    """Fixture for a django db user"""

    password = "fake"
    user = get_user_model().objects.create_user(username="testuser",
                                                password=password)
    user.unhashed_password = password

    def fin():
        user.delete()
    request.addfinalizer(fin)

    return user


@pytest.fixture
def test_group(request, db):
    """Fixture for a django test group"""
    group = Group.objects.create(name="test_group")

    def fin():
        group.delete()
    request.addfinalizer(fin)

    return group


@pytest.fixture
def test_permission(request, db):
    """Fixture for a django test permission"""

    content_type = ContentType.objects.get_for_model(get_user_model())
    permission = Permission.objects.create(codename="test_permission",
                                           name="Test Permission",
                                           content_type=content_type)

    def fin():
        permission.delete()
    request.addfinalizer(fin)

    return permission


@pytest.fixture
def patch_authenticate_success(request, db, monkeypatch):
    """Fixture to patch successful authentication"""

    monkeypatch.setattr("requests.sessions.Session.request",
                        lambda *args, **kwargs:
                        namedtuple("Response", ['status_code'])(200))
    monkeypatch.setattr("freeipa_auth.freeipa_utils."
                        "FreeIpaSession._get_user_data",
                        lambda *args: None)


@pytest.fixture
def patch_authenticate_fail(request, db, monkeypatch):
    """Fixture patch a failed authentication"""

    monkeypatch.setattr("requests.sessions.Session.request",
                        lambda *args, **kwargs:
                        namedtuple("Response", ['status_code'])(401))


@pytest.fixture
def patch_remote_user_groups(request, db, monkeypatch):
    """Fixture to patch remote user groups"""

    monkeypatch.setattr("freeipa_auth.freeipa_utils.FreeIpaSession.groups",
                        ["admin", "test_group",
                         "test_permission"])


@pytest.fixture
def settings(request, db):
    """Fixture to allow for setting overrides per test case"""

    def override(**kwargs):
        for k, v in kwargs.items():
            setattr(django_settings, k, v)

        def fin():
            for k in kwargs:
                delattr(django_settings, k)

        request.addfinalizer(fin)

    django_settings.override = override
    return django_settings


@pytest.fixture
def liveserver_username(request, db):
    """Fixture to use a liveserver username"""
    return input("username: ")


@pytest.fixture
def liveserver_password(request, db):
    """Fixture to use a liveserver password"""
    return getpass.getpass("password: ")


@pytest.fixture
def liveserver(request, db):
    """Fixture to use a liveserver for testing (passed in from command line)"""
    return request.config.getoption('liveserver')



from django.test import Client
from django .conf import settings
from django.contrib.auth.models import User
import pytest


class TestFreeIpaBackendAuth(object):

    client = Client()
    username = "dummy_freeipa_username"
    password = "dummy_freeipa_password"

    def test_login(self, settings, patch_authenticate_success):
        """Test succesful login"""

        settings.override(FREEIPA_AUTH_AUTHORIZE_ALL_USERS=True)
        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in
        user = User.objects.get(username=self.username)

        # No permissions on basic login
        assert not user.is_staff
        assert not user.is_superuser

    def test_logout(self, patch_authenticate_success):
        """Test successful logout"""

        settings.override(FREEIPA_AUTH_AUTHORIZE_ALL_USERS=True)
        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in
        logged_in = self.client.logout()
        assert not logged_in

    def test_update_user_groups(self, test_group, test_permission, settings,
                                patch_authenticate_success, patch_remote_user_groups):
        """Test that user groups are update on first time login"""

        settings.override(FREEIPA_AUTH_UPDATE_USER_GROUPS=True,
                          FREEIPA_AUTH_UPDATE_USER_PERMISSIONS_BY_GROUP=True,
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"]})

        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in
        user = User.objects.get(username=self.username)

        # Since the "is_staff" flag exists in the settings
        # the user will be a staff member
        assert user.is_staff

        # No mapping was provided for the superuser
        # flag so user will not be a superuser
        assert not user.is_superuser

        # The user is part of "test_group" and also has
        # "test_permission" on the freeipa server so they
        # will updated in django as well
        assert test_group in user.groups.all()
        assert test_permission in user.user_permissions.all()

    def test_update_user_groups_with_prefix(self, test_group, test_permission, monkeypatch, settings,
                                            patch_authenticate_success, patch_remote_user_groups):
        """Test that user groups are mapped with a required group prefix"""

        settings.override(FREEIPA_AUTH_UPDATE_USER_GROUPS=True,
                          FREEIPA_AUTH_UPDATE_USER_PERMISSIONS_BY_GROUP=True,
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"]},
                          FREEIPA_AUTH_REQUIRE_GROUP_PREFIX="foo.django.group.",
                          FREEIPA_AUTH_REQUIRE_PERMISSION_PREFIX="foo.django.permission.")

        # Patch user groups on freeipa to have the required prefix for mapping
        monkeypatch.setattr("freeipa_auth.freeipa_utils.FreeIpaSession.groups",
                            [settings.FREEIPA_AUTH_REQUIRE_GROUP_PREFIX + "admin",
                             settings.FREEIPA_AUTH_REQUIRE_GROUP_PREFIX + "test_group",
                             settings.FREEIPA_AUTH_REQUIRE_PERMISSION_PREFIX + "test_permission"])

        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in

        # Assert that the user in the mapped django
        # groups and has the mapped permission
        user = User.objects.get(username=self.username)
        assert user.is_staff
        assert test_group in user.groups.all()
        assert test_permission in user.user_permissions.all()

    def test_update_user_attrs(self, monkeypatch, settings,
                               patch_authenticate_success, patch_remote_user_groups):
        """Test that user attrs are updated on first time login"""

        # Mock user data from freeipa
        monkeypatch.setattr("freeipa_auth.freeipa_utils.FreeIpaSession._get_user_data",
                            lambda *args: {"givenname": ['Jerry'], 'mail': ['jerry@django.com']})

        # Set user attrs map and group flags in settings
        settings.override(FREEIPA_AUTH_USER_ATTRS_MAP={"first_name": "givenname", "email": "mail"},
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"]})
        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in

        # Assert that user attrs are mapped and saved
        user = User.objects.get(username=self.username)
        assert user.first_name == "Jerry"
        assert user.email == 'jerry@django.com'

    def test_always_update_user(self, settings, monkeypatch,
                                patch_authenticate_success, patch_remote_user_groups):
        """Test that user is always updated on subsequent logins if set to True in settings"""

        settings.override(FREEIPA_AUTH_ALWAYS_UPDATE_USER=True,
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"],
                                                            'is_superuser': ['superuser']})

        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in
        user = User.objects.get(username=self.username)

        # Assert that initially user is not a superuser
        assert not user.is_superuser
        logged_in = self.client.logout()
        assert not logged_in

        # Patch user groups on freeipa to have the superuser flag
        monkeypatch.setattr("freeipa_auth.freeipa_utils.FreeIpaSession.groups",
                            ["admin", "test_group", "superuser", "test_permission"])

        # Login again
        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in

        # User should now be superuser since
        # FREEIPA_AUTH_ALWAYS_UPDATE_USER is set to True in settings
        user = User.objects.get(username=self.username)
        assert user.is_superuser

    def test_authorize_all_users(self, settings, patch_authenticate_success):
        """
        Test that user is authorized to be staff by default
        if FREEIPA_AUTH_AUTHORIZE_ALL_USERS setting is set to True
        """

        # No user groups are needed to authorize basic login
        settings.override(FREEIPA_AUTH_AUTHORIZE_ALL_USERS=True)
        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in

    def test_not_authorize_all_users(self, settings, patch_authenticate_success):
        """
        Test that user is authorized to be staff by default
        if FREEIPA_AUTH_AUTHORIZE_ALL_USERS setting is set to True
        """

        settings.override(FREEIPA_AUTH_AUTHORIZE_ALL_USERS=False,
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"],
                                                            'is_superuser': ['superuser']})
        logged_in = self.client.login(username=self.username, password=self.password)
        # User cannot login since they do not have any user group flags on remote server
        assert not logged_in

    def test_not_authorize_all_users_with_groups(self, settings,
                                                      patch_authenticate_success, patch_remote_user_groups):
        """
        Test that user is authorized to be staff by default
        if FREEIPA_AUTH_AUTHORIZE_ALL_USERS setting is set to True
        """

        # No user groups are needed to authorize basic login
        settings.override(FREEIPA_AUTH_AUTHORIZE_ALL_USERS=False,
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"],
                                                            'is_superuser': ['superuser']}
                          )
        logged_in = self.client.login(username=self.username, password=self.password)
        # User can login when AUTHORIZE_ALL_USERS is set to False
        # and are members of groups that are user flags
        assert logged_in

    def test_no_update_user(self, settings, monkeypatch,
                            patch_authenticate_success, patch_remote_user_groups):
        """Test that user is not updated on subsequent logins if set to False in settings"""

        settings.override(FREEIPA_AUTH_ALWAYS_UPDATE_USER=False,
                          FREEIPA_AUTH_USER_FLAGS_BY_GROUP={"is_staff": ["admin"],
                                                            'is_superuser': ['superuser']})

        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in

        user = User.objects.get(username=self.username)

        # Assert that initially user is not a superuser
        assert not user.is_superuser
        logged_in = self.client.logout()
        assert not logged_in

        # Patch user groups on freeipa to have the superuser flag
        monkeypatch.setattr("freeipa_auth.freeipa_utils.FreeIpaSession.groups",
                            ["admin", "test_group", "superuser", "test_permission"])

        # Login again
        logged_in = self.client.login(username=self.username, password=self.password)
        assert logged_in

        # User should still not be superuser since FREEIPA_AUTH_ALWAYS_UPDATE_USER is set to False
        user = User.objects.get(username=self.username)
        assert not user.is_superuser

    def test_invalid_credentials(self, patch_authenticate_fail):
        """Test that no django user is created when login credentials are invalid"""

        logged_in = self.client.login(username=self.username, password=self.password)
        assert not logged_in

        # User should not be in the database
        with pytest.raises(User.DoesNotExist):
            User.objects.get(username=self.username)

    def test_classic_django_auth(self, test_user):
        """Test that classic django auth is still the main authentication backend"""

        # Here we can see we do not need to patch the freeipa response
        # since it does not reach the freeipa backend auth when a
        # user uses the django app login credentials
        logged_in = self.client.login(username=test_user.username, password=test_user.unhashed_password)
        assert logged_in

    @pytest.mark.skip(reason="Don't want to automate remote server calls")
    def test_login_live(self, settings, liveserver_username, liveserver_password, liveserver):
        """Test succesful login on live server"""

        settings.override(FREEIPA_AUTH_SERVER=liveserver,
                          FREEIPA_AUTH_SSL_VERIFY=False)

        logged_in = self.client.login(username=liveserver_username, password=liveserver_password)

        assert logged_in
        assert User.objects.get(username=liveserver_username)

    @pytest.mark.skip(reason="Don't want to automate remote server calls")
    def test_login_live_failover(self, settings, liveserver_username, liveserver_password, liveserver):
        """
        Test authentication falls back to failover
        server if there is a connection error on main server
        """

        settings.override(FREEIPA_AUTH_SERVER="test.fake-site.com",
                          FREEIPA_AUTH_FAILOVER_SERVER=liveserver,
                          FREEIPA_AUTH_SSL_VERIFY=False)

        logged_in = self.client.login(username=liveserver_username, password=liveserver_password)

        # Client will authenticate on failover
        # server and be logged in on the django app
        assert logged_in
        assert User.objects.get(username=liveserver_username)



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-freeipa-auth',
    version='0.1.dev7',
    packages=find_packages(),
    include_package_data=True,
    license='BSD License',
    description='A simple django freeipa rpc authentication backend app with a simple server failover solution.',
    long_description=README,
    tests_require=['pytest', 'pytest-django>=2.9.1,<=3.1.2', 'requests', 'Django >= 1.8.0'],
    install_requires=['requests', 'Django >= 1.8.0'],
    extras_require={
        'security': ['pyOpenSSL >= 0.14', 'cryptography>=1.3.4', 'idna>=2.0.0'],
    },
    author='Jerry Aujla',
    author_email='jerryaujla1@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
